
#ifndef __COLLECTION__
#define __COLLECTION__

#include "utils/CString.h"

class TTable;
class TTableIterator;

/** TTableItem *************************************************************/
// base class for items in TTable

class TTableItem {
        friend class TTable;
        friend class TTableIterator;
public:
        TTableItem( const CString& name )
            : mName(name), mPrev(0), mNext(0), mHashPrev(0), mHashNext(0)
            { mHashCode = mName.hashCode(); };
        virtual ~TTableItem() { };

        const CString&  getName()       const { return mName; };
        long            getHashCode()   const { return mHashCode; };

        TTableItem     *getNext()       const { return mNext; };

private:
        CString     mName;
        long        mHashCode;
        TTableItem *mPrev, *mNext;   // list
        TTableItem *mHashPrev, *mHashNext; // bucket list
};

/** TTable *****************************************************************/
// table (list & hash) of TTableItems

/* manages order of TTableItems (via list - get with iterator), also
   internally organizes hash table (by item names) for fast searches */
/* once put into a table, an item is owned by it */

class TTable {
        friend class TTableIterator;
public:
        TTable( long hashSize ); // size: 2^hashSize
        ~TTable();

        void        addItem( TTableItem *item );
        TTableItem *findItem( const CString& name ) const;

        long        getSize() const { return mSize; }
        TTableItem *getFirst() const { return mFirst; }

        TTableIterator items() const;

private:
        struct TBucket {
            TTableItem  *mFirst;
            TTableItem  *mLast;
        };

        long            mHashSize;  // # of hash buckets
        TBucket        *mHash;      // bucket array

        TTableItem     *mFirst;     // first in list
        TTableItem     *mLast;      // last in list
        long            mSize;      // # of items
};

/** TTableIterator *********************************************************/
// iterator for TTable list:

// do like this:
/*      it = table.items();
        while (it.hasNext()) {
            item = it.next();
            // do smth with item
        }
*/

class TTableIterator {
public:
        TTableIterator( const TTable& table );
        bool        hasNext() const;
        TTableItem *next();
private:
        const TTable&   mTable;
        TTableItem     *mCurrent;
};

// inlines
inline TTableIterator::TTableIterator( const TTable& table )
    : mTable(table), mCurrent( table.mFirst ) { };
inline bool TTableIterator::hasNext() const
    { return (mCurrent != NULL); }
inline TTableItem *TTableIterator::next()
    { TTableItem *i = mCurrent; mCurrent = mCurrent->mNext; return i; }
inline TTableIterator TTable::items() const { return TTableIterator(*this); }

#endif


