// sintakses medis - parsinimo metu daro simboliu ir tipu lenteles ir sintakses
// medi. medyje - visi sakiniai programos (pati deela).

// funkcijos eina taip: calleris steke padaro vietos rezultatui, tada sudeda
// visus parametrus (normalia tvarka), tada irasho objekto adresa (__this),
// tada daro call (irasho grizhimo adresa). grizhimo adresas - hackas, ji
// galima net keisti (jis yra lokalus kint. vardu "__return_addr". funkcija
// dar ishsiskiria steke vietos lokaliems, priesh isheidama juos sunaikina.
// pakeicia grazinama rezulta (su return arba per kintamaji "__return_val").
// tada kaip tik steko virshuje lieka grizhimo adr - tai ir daro ret (ir
// ishima adr). calleris turi ishimti (arba ne) parametrus.

// adresai yra taip: visi adresai imami kaip ptr ish steko. steke ptr.
// gali buti suformuojamas: push base ptr. ir tada dar pridedant/atimant
// ish jo galutinis. base ptr gali buti: push referenca arba push
// frame-ptr, t.y. steke esantiems lokaliems pasiekti.

#ifndef __SYNTAXTREE__
#define __SYNTAXTREE__

#include "tclass.h"
#include "opcodes.h"


/** ABSTRAKCIJOS ***********************************************************/

struct TIntInsn;

// abstrakti medzhio shaka
class TNode {
public:
        virtual void debug(long n) const;
        // susioptimizuoja (gali pakeisti save)
        virtual TNode *optimize()       { return this; };
        // nusioptimizuoja, palieka tik tuos, kur su pashaliniais efektais
        virtual TNode *remove()         { return this; };
        // nodo grazinamos reiksmes tipas
        virtual TType *type() const     = 0;
        // susikodina i tarpini koda
        virtual TIntInsn *code() const;
        // destruktina save
        virtual ~TNode()                { };
        // destruktina ir savo vaikus
        virtual void del()              { delete this; };

		// returns n if n matches t type. coerces n if could match.
		// supports multi-expr n
		static TNode *matchType( TType *t, TNode *n, const char *descr );
};

// abstrakti shaka - nieko negrazina (for, if, ...)
class TVoidNode : public TNode {
public:
        TType *type() const { return TypeVoid; };
};
// abstrakti shaka - kazka grazina (expressionai visokie)
class TRetNode : public TNode {
};
// dalyku listas
class TSeqNode : public TVoidNode {
protected:
        TNode *list; // kiti dalykai
        TNode *node; // vienas dalykas
public:
        TSeqNode(TNode *l, TNode *n) { list = l; node = n; };
        void del() { list->del(); node->del(); delete this; };
};
// tikrinimas+sakinys (if, while, do)
class TCondStatNode : public TVoidNode {
protected:
        TNode *cond;
        TNode *stat;
public:
        TCondStatNode(TNode *c, TNode *s) { cond = c; stat = s; };
        void del() { cond->del(); stat->del(); delete this; };
        void debug(long n) const;
        TNode *optimize();
};

/** SAKINIAI ***************************************************************/

// tuscias sakinys
class TEmptyNode : public TVoidNode {
public:
        TEmptyNode() { };
        void debug(long n) const;
        TType *type() const { return TypeVoid; };
};

// sakiniu listas
class TListNode : public TSeqNode {
public:
        TListNode(TNode *l, TNode *n) : TSeqNode(l,n) { };
        void debug(long n) const;
        TNode *optimize();
        TNode *remove();
        TIntInsn *code() const;
};

// IF sakinys
class TIfNode : public TCondStatNode {
public:
        TIfNode(TNode *c, TNode *s) : TCondStatNode(c,s) { };
        void debug(long n) const;
        TIntInsn *code() const;
};

// IF (su ELSE) sakinys
class TIfElseNode : public TIfNode {
protected:
        TNode *stat2;
public:
        TIfElseNode(TNode *c, TNode *s, TNode *s2) : TIfNode(c,s) { stat2 = s2; };
        void del() { stat2->del(); cond->del(); stat->del(); delete this; };
        void debug(long n) const;
        TIntInsn *code() const;
};

// trinaris operatorius A ? B : C
class TTernaryNode : public TIfElseNode {
public:
        TTernaryNode(TNode *c, TNode *s, TNode *s2) : TIfElseNode(c,s,s2) { };
        void debug(long n) const;
        TIntInsn *code() const;
        TNode *optimize();
        TType *type() const;
};

// WHILE sakinys
class TWhileNode : public TCondStatNode {
public:
        TWhileNode(TNode *c, TNode *s) : TCondStatNode(c,s) { };
        void debug(long n) const;
        TIntInsn *code() const;
};

// DO/UNTIL sakinys
class TDoNode : public TCondStatNode {
public:
        TDoNode(TNode *c, TNode *s) : TCondStatNode(c,s) { };
        void debug(long n) const;
        TIntInsn *code() const;
};

// FOR sakinys
class TForNode : public TVoidNode {
        TNode *init;
        TNode *cond;
        TNode *loop;
        TNode *stat;
public:
        TForNode(TNode *i, TNode *c, TNode *l, TNode *s)
                { init = i; cond = c; loop = l; stat = s; };
        void del() { init->del(); cond->del(); loop->del(); stat->del(); delete this; };
        void debug(long n) const;
        TNode *optimize();
        TIntInsn *code() const;
};

// RETURN sakinys
class TReturnNode : public TVoidNode {
public:
        void debug(long n) const;
        TIntInsn *code() const;
};

// DELETE sakinys
class TDeleteNode : public TVoidNode {
        TNode *lval;
public:
        TDeleteNode(TNode *l) : lval(l) { };
        void del() { lval->del(); delete this; };
        void debug(long n) const;
        TIntInsn *code() const;
};

/** BINARINES OP ***********************************************************/

// operacija su 2 operandais op1 ir op2
class TBinaryNode : public TRetNode {
        typedef float (*TOptiFunc)(float, float);
        struct TParams {
                const char *name; // vardas (debug)
                EIntOpcode  opi;  // opkodas int
                EIntOpcode  opf;  // opkodas flt
                TOptiFunc   opti; // optimiz. f-ja
                bool        flt;  // ar leidziama flt
                bool        ref;  // ar leidziama ref
                bool        lng;  // ar rez. long, neprikl. nuo arg.
        };
protected:
        TNode      *op1;  // operandai
        TNode      *op2;
        TParams    *p;    // parametrai tam f-jos tipui
        static TParams table[];
public:
        enum EMode {
                eADD, eSUB, eMUL, eDIV, eMOD, eSHL, eSHR, eOR, eXOR, eAND,
                ebOR, ebAND, ebE, ebNE, ebL, ebG, ebLE, ebGE
        };
        TBinaryNode(EMode m, TNode *o1, TNode *o2)
            { p = &table[m]; op1 = o1; op2 = o2; };
        void del()
            { op1->del(); op2->del(); delete this; };
        TType *type() const;
        void debug(long n) const;
        TNode *optimize();
        TNode *remove();
        TIntInsn *code() const;
};

/** UNARINES OP ************************************************************/

// operacija su 1 operandu op
class TUnaryNode : public TRetNode {
        typedef float (*TOptiFunc)(float);
        struct TParams {
                const char *name; // vardas (debug)
                EIntOpcode  opi;  // opkodas int
                EIntOpcode  opf;  // opkodas flt
                TOptiFunc   opti; // optimiz. f-ja
                bool        flt;  // ar leidziama flt
                bool        ref;  // ar leidziama ref
                bool        lng;  // ar rez. long, neprikl. nuo arg.
        };
protected:
        TNode      *op;   // operandas
        TParams    *p;    // parametrai tam f-jos tipui
        static TParams table[];
public:
        enum EMode {
                epreINC, epreDEC, epostINC, epostDEC,
                eMINUS, ebNOT, eINV
        };
        TUnaryNode(EMode m, TNode *o)
            { p = &table[m]; op = o; };
        void del()
            { op->del(); delete this; };
        TType *type() const;
        void debug(long n) const;
        TNode *optimize();
        TNode *remove();
        TIntInsn *code() const;
};

/** FUNKCIJU KVIETIMAS *****************************************************/

// funkcijai kviesti turim visu argumentu shaku sarasha
struct TArgs {
        TArgs(TNode *e) : expr(e), next(NULL) { };
        ~TArgs() { expr->del(); };
        TNode *expr;    // argumento nodas
        TArgs *next;    // sekantis arg
};

// funkcijos kvietimas
class TFuncNode : public TRetNode {
        TNode     *lval;        // objektas
        TFunction *func;        // kuri f-ja
        TArgs     *args;        // argumentai
        long       count;       // arg kiekis
        long       super;       // jei >= 0 - kurios klases
public:
        TFuncNode(TNode *l, TFunction *f, TArgs *a, long s);
        ~TFuncNode();
        void del() { lval->del(); delete this; };
        void debug(long n) const;
        TNode *optimize();
        TType *type() const   { return func->Type(); };
        TIntInsn *code() const;
};

// native funkcijos kvietimas
class TNativeNode : public TVoidNode {
        long slot; // kuris native slotas
public:
        TNativeNode(long s) : slot(s) { };
        void debug(long n) const;
        TType *type() const { return TypeVoid; };
        TIntInsn *code() const;
};

/** Multi-expr isaiskos *************************************************/

// { expr, expr, ... }
class TMultiExprNode : public TRetNode {
        TArgs     *exprs;       // israiskos
        long       count;       // israisku kiekis
public:
        TMultiExprNode(TArgs *a);
        ~TMultiExprNode();

		TArgs  *Exprs() const { return exprs; }
		long	Count() const { return count; }
		long	TotalSize() const;

        void debug(long n) const;
        TType *type() const   { return TypeVoid; };
        TIntInsn *code() const;
};

/** STD. KONVERSIJOS *******************************************************/

class TCoerceNode : public TRetNode {
protected:
        TNode *op;
public:
        TCoerceNode(TNode *o) { op = o; };
        TNode *remove();
        void   del()          { op->del(); delete this; };
};

// float -> long
class T2Long : public TCoerceNode {
public:
        T2Long(TNode *o) : TCoerceNode(o) { };
        void debug(long n) const;
        TType *type() const { return TypeLong; };
        TNode *optimize();
        TIntInsn *code() const;
};
// long -> float
class T2Float : public TCoerceNode {
public:
        T2Float(TNode *o) : TCoerceNode(o) { };
        void debug(long n) const;
        TType *type() const { return TypeFloat; };
        TNode *optimize();
        TIntInsn *code() const;
};

/** LVALUE *****************************************************************/

// nodai, kuriu rezultatas - rodykle i kazka
class TLvalNode : public TRetNode {
};

// prie steke esanchio referenco prideda kazhkiek (masyvams)
class TOffsetNode : public TLvalNode {
        TNode *lval;    // sukuria steke ref
        TNode *expr;    // prideda
public:
        TOffsetNode(TNode *l, TNode *e) : lval(l), expr(e) { };
        void del() { lval->del(); expr->del(); delete this; };
        void debug(long n) const;
        TNode *optimize();
        TNode *remove();
        TType *type() const   { return lval->type(); };
        TIntInsn *code() const;
};
// prie steke esanchio referenco prideda kazhkiek (memberiams)
class TMemberNode : public TLvalNode {
        TNode *lval;    // sukuria steke ref
        TSymbol *sym;
public:
        TMemberNode(TNode *l, TSymbol *s) : lval(l), sym(s) { };
        void del()      { lval->del(); delete this; };
        void debug(long n) const;
        TNode *remove();
        TType *type() const   { return sym->Type(); };
        TIntInsn *code() const;
};
// ipushina i steka frame-ptr (steko pradzia), static-ptr (static) arba "this"
class TBaseRefNode : public TLvalNode {
        TType *typ;     // tipas (this); void (frame-ptr; static-ptr)
        enum { E_THIS, E_FRAME, E_STATIC } mode;
public:
        TBaseRefNode(TType *t)
            : typ(t) { mode = E_THIS; }; // "this"
        TBaseRefNode(bool s = false)
            : typ(TypeRef) { mode = s ? E_STATIC:E_FRAME; }; // "frame/static"
        void del() { delete this; };
        void debug(long n) const;
        TNode *remove()         { del(); return new TEmptyNode(); };
        TType *type() const     { return typ; };
        TIntInsn *code() const;
};
// NEW TIPAS; palieka steke ref i objekta
class TNewNode : public TLvalNode {
        TClass *cla;
public:
        TNewNode(TClass *c) : cla(c) { };
        void      debug(long n) const;
        TType    *type() const  { return cla->Type(); };
        TIntInsn *code() const;
};
// ishima ish steko referenca, pagal ji ipushina i steka dalyka.
// lval gali tureti shal. efekta
class TIdentNode : public TRetNode {
        TNode *lval; // sukuria steke ref
public:
        TIdentNode(TNode *l) : lval(l) { };
        void del() { lval->del(); delete this; };
        void debug(long n) const;
        TNode *remove(); // netinka ish terminalnode
        TType *type() const     { return lval->type(); };
        TIntInsn *code() const;
        TNode *Lval() const     { return lval; };
};
// grazina bool - "(dynamic_cast <cla> (lval)) != NULL"
class TIsNode : public TLvalNode {
        TNode *lval;
        TClass *cla;
public:
        TIsNode(TNode *l, TClass *c) : lval(l), cla(c) { };
        void del() { lval->del(); delete this; };
        void      debug(long n) const;
        TNode    *remove();
        TType    *type() const  { return TypeLong; };
        TIntInsn *code() const;
};

/** TERMINATORIAI (boo!) ***************************************************/

// jei jie nieko neitakoja, tai galima ishimti (tik identnode nevisada),
// taigi turi remove()

// galutiniai nodai, nuo nieko kito nepriklauso, shalutiniu efektu neturi
class TTerminalNode : public TRetNode {
public:
        TNode *remove()         { del(); return new TEmptyNode(); };
};
// skaitmenine konstanta
class TNumericNode : public TTerminalNode {
protected:
        float val;
public:
        float Val() const { return val; };
};

// long konstanta
class TNumberNode : public TNumericNode {
public:
        TNumberNode(long v) { val = v; };
        void debug(long n) const;
        TType *type() const    { return TypeLong; };
        TIntInsn *code() const;
};
// float konstanta
class TFloatNode : public TNumericNode {
public:
        TFloatNode(float v) { val = v; };
        void debug(long n) const;
        TType *type() const    { return TypeFloat; };
        TIntInsn *code() const;
};

// string konstanta
class TStringNode : public TTerminalNode {
        long offset;
public:
        TStringNode(long o) : offset(o) { };
        void debug(long n) const;
        TType *type() const    { return TypeStr; };
        TIntInsn *code() const;
};

/** PRISKYRIMO OP **********************************************************/

// priskyrimas - pirma sukuriam steke expr, tada lval, tada i lval irashom
// expr. expr steke lieka (kad a=b=c=d galima butu). priskyrimo tevai
// turi pasirupinti jo ishemimu, jei nebereikia

// remove() neturi, nes yra sakinio efektas (priskiria)

// paprastas priskyrimas '='
class TAssignNode : public TRetNode {
protected:
        TNode *expr; // sukuria steke expr
        TNode *lval; // sukuria steke lval
public:
        TAssignNode(TNode *l, TNode *e) { lval = l; expr = e; };
        void del() { lval->del(); expr->del(); delete this; };
        TType *type() const { return lval->type(); };
        void debug(long n) const;
        TNode *optimize();
        TIntInsn *code() const;
};

// aritmetinis: += -= ...
class TAsArithNode : public TAssignNode {
        struct TParams {
                const char *name; // vardas (debug)
                EIntOpcode  opi;  // opkodas int
                EIntOpcode  opf;  // opkodas flt
                bool        flt;  // ar leidziama flt
        };
protected:
        TParams    *p;    // parametrai tam f-jos tipui
        static TParams table[];
public:
        enum EMode {
                eaMUL, eaDIV, eaMOD, eaADD, eaSUB, eaSHL, eaSHR,
                eaOR, eaXOR, eaAND,
        };
        TAsArithNode(EMode m, TNode *l, TNode *e) : TAssignNode(l, e)
            { p = &table[m]; };
        TType *type() const;
        void debug(long n) const;
        TNode *optimize();
        TIntInsn *code() const;
};

/** IDENTIFIKACIJAI NODU ***************************************************/

// ar saka yra tam tikro tipo
static inline TEmptyNode     *isEmpty    (TNode *n) { return dynamic_cast<TEmptyNode*>    (n); };
static inline TRetNode       *isRet      (TNode *n) { return dynamic_cast<TRetNode*>      (n); };
static inline TMultiExprNode *isMultiExpr(TNode *n) { return dynamic_cast<TMultiExprNode*>(n); };
static inline TBinaryNode    *isBinary   (TNode *n) { return dynamic_cast<TBinaryNode*>   (n); };
static inline TUnaryNode     *isUnary    (TNode *n) { return dynamic_cast<TUnaryNode*>    (n); };
static inline TTerminalNode  *isTerminal (TNode *n) { return dynamic_cast<TTerminalNode*> (n); };
static inline TLvalNode      *isLval     (TNode *n) { return dynamic_cast<TLvalNode*>     (n); };
static inline TListNode      *isList     (TNode *n) { return dynamic_cast<TListNode*>     (n); };
static inline TNumericNode   *isNumeric  (TNode *n) { return dynamic_cast<TNumericNode*>  (n); };
static inline TBaseRefNode   *isBaseRef  (TNode *n) { return dynamic_cast<TBaseRefNode*>  (n); };

#endif

